#include "CGround.h"

CGround::CGround(IrrlichtDevice * pDevice,int width,int height)
    : m_width(width), m_height(height), m_node(NULL), m_selector(NULL)
{
    // Our first step is to determine how many vertices
    // we will need to create a mesh that is width x height
    //
    //  *----*----*----*  As you can see, we will need exactly
    //  |    |    |    |  (width + 1) * (height + 1) vertices
    //  *----*----*----*  and (6 * width * height) indices for
    //  |    |    |    |  our index buffer
    //  *----*----*----*

    int iNumVertices = (width + 1) * (height + 1);
    S3DVertex * vertices = new S3DVertex[iNumVertices];
    memset(vertices,0,sizeof(S3DVertex) * iNumVertices);  // Set everything to 0

    // Now loop through the vertices and fill them with the appropriate
    // values

    for(int i=0;i<=height;++i)
    {
        for(int j=0;j<=width;++j)
        {
            int iIndex = (i*(width + 1)) + j;

            // Store this vertex's coords and color
            vertices[iIndex].Pos.X = j * 2.0f;
            vertices[iIndex].Pos.Y = 0.0f;
            vertices[iIndex].Pos.Z = i * 2.0f;
            vertices[iIndex].Color.color = 0xFFFFFFFF;

            // Also store the texture UV values
            vertices[iIndex].TCoords.X = i;
            vertices[iIndex].TCoords.Y = j;
        }
    }

    // Now we need to create the index buffer
    int iNumIndices = 6 * width * height;
    u16 * indices = new u16[iNumIndices];

    for(int i=0;i<height;++i)
    {
        for(int j=0;j<width;++j)
        {
            int iIndex = ((i*width) + j) * 6;

            // Now store the values in the array
            int tmp_offset = j + (i * (width + 1));

            indices[iIndex + 0] = tmp_offset + 1;
            indices[iIndex + 1] = tmp_offset;
            indices[iIndex + 2] = tmp_offset + width + 1;
            indices[iIndex + 3] = tmp_offset + 1;
            indices[iIndex + 4] = tmp_offset + width + 1;
            indices[iIndex + 5] = tmp_offset + width + 2;
        }
    }

    // Create the mesh buffer
    SMeshBuffer * buffer = new SMeshBuffer();
    buffer->append(vertices,iNumVertices,indices,iNumIndices);

    // Load the grass texture
    SMaterial mat;
    mat.setTexture(0,pDevice->getVideoDriver()->getTexture(DATA_DIR "/textures/grass.jpg"));
    buffer->Material = mat;

    // Create the SMesh
    SMesh * mesh = new SMesh();
    mesh->addMeshBuffer(buffer);

    mesh->recalculateBoundingBox();

    // Drop the ref. count of the buffer
    buffer->drop();

    // ...and free the memory held by the data
    delete [] indices;
    delete [] vertices;

    // Add this to the scene
    m_node = pDevice->getSceneManager()->addMeshSceneNode(mesh);

    // In order to make the ground look good at a
    // distance, we need to enable anisotropic filtering
    m_node->setMaterialFlag(EMF_ANISOTROPIC_FILTER,true);

    // Lastly, we create a triangle selector for this
    // mesh so that when we want to place things, we can
    // determine where the user is pointing the cursor
    m_selector = pDevice->getSceneManager()->createTriangleSelector(mesh, m_node);
    m_node->setTriangleSelector(m_selector);

    // Drop the ref. count of the mesh
    mesh->drop();
}

CGround::~CGround()
{
    m_selector->drop();

    // Do NOT drop the scene node pointer
    // instead remove it
    m_node->remove();
}

bool CGround::Intersection(IrrlichtDevice * device, line3df ray, int & x_inter, int & z_inter)
{
    vector3df out_p;
    triangle3df out_t;
    const ISceneNode * out_s;

    bool bCollision = device->getSceneManager()->getSceneCollisionManager()->
                      getCollisionPoint(ray, m_selector, out_p, out_t, out_s);

    if(bCollision)  // one was found
    {
        // See if we can figure out which one
        // by looping through the points on the
        // triangle and getting the minimum value
        // for each dimension.
        float min_x = (out_t.pointA.X < out_t.pointB.X)?out_t.pointA.X:out_t.pointB.X;
        min_x = (out_t.pointC.X < min_x)?out_t.pointC.X:min_x;

        float min_z = (out_t.pointA.Z < out_t.pointB.Z)?out_t.pointA.Z:out_t.pointB.Z;
        min_z = (out_t.pointC.Z < min_z)?out_t.pointC.Z:min_z;

        // The track coords for this piece can be found by dividing the
        // min_x and min_z values by 2
        x_inter = min_x / 2.0f;
        z_inter = min_z / 2.0f;

        return true;
    }
    else
        return false;
}
