#include "CPlayingState.h"

// Definitions
#define PI2 (PI/2.0f)

// Global ID enum
enum GUI_WINDOW_ID { // Toolbar buttons
                     TOOLBAR_EDIT_TRACK,
                     TOOLBAR_ADD_TRAIN,

                     // Edit track dialog
                     EDITTRACK_STRAIGHT,
                     EDITTRACK_CURVED,

                     // Add train dialog
                     ADDTRAIN_ADD };


CPlayingState::CPlayingState(CStateManager * pManager)
    : CGameState(pManager), m_camera(NULL), m_toolbar(NULL), m_mode(IM_SELECT), m_ground(GetDevice(),25,25),
      m_sky(GetDevice(),25,25), m_track_manager(GetDevice(),25,25), m_train_manager(&m_track_manager),
      m_straight_ghost_node(NULL), m_curve_ghost_node(NULL), m_engine_ghost_node(NULL), m_dialog_edittrack(NULL), m_dialog_addtrain(NULL),
      m_previous_time(0), m_rotation(0.0f), m_radius(3.0f),  m_up(false), m_down(false), m_left(false), m_right(false),
      m_r_down(false), m_x_diff(0.0f), m_y_diff(0.0f), m_x_cur(0), m_y_cur(0), m_x_prev(0), m_y_prev(0),
      m_current_class(TC_STRAIGHT), m_current_track(TT_VERTICAL_STRAIGHT)
{
    // Start by setting the ambient lighting
    // and initializing the camera
    GetDevice()->getSceneManager()->setAmbientLight(SColorf(0.7,0.7,0.7,1));
    m_camera = GetDevice()->getSceneManager()->addCameraSceneNode();

    // Set the camera's rotation, target, FOV, etc.
    m_camera->bindTargetAndRotation(true);
    m_camera->setPosition(vector3df(30,4,2));
    m_camera->setTarget(vector3df(30,2,5));
    m_camera->setFOV(PI / 3.5f);
    m_camera->setAspectRatio(1.66f);

    // We need to set m_previous_time to a proper value
    m_previous_time = GetDevice()->getTimer()->getTime();

    // Create the toolbar
    CreateToolbar();

    // Load the ghost nodes
    IMesh * straight_ghost = GetDevice()->getSceneManager()->getMesh(DATA_DIR "/models/track_straight.b3d");
    m_straight_ghost_node = GetDevice()->getSceneManager()->addMeshSceneNode(straight_ghost);

    IMesh * curve_ghost = GetDevice()->getSceneManager()->getMesh(DATA_DIR "/models/track_small_curve.b3d");
    m_curve_ghost_node = GetDevice()->getSceneManager()->addMeshSceneNode(curve_ghost);

    IMesh * engine_ghost = GetDevice()->getSceneManager()->getMesh(DATA_DIR "/models/simple_engine_ghost.b3d");
    m_engine_ghost_node = GetDevice()->getSceneManager()->addMeshSceneNode(engine_ghost);

    // Make them all invisible
    m_straight_ghost_node->setVisible(false);
    m_curve_ghost_node->setVisible(false);
    m_engine_ghost_node->setVisible(false);
}

CPlayingState::~CPlayingState()
{
    // Remove the camera
    m_camera->remove();
}

bool CPlayingState::OnEvent(const SEvent & event)
{
    // Switch on the event type
    switch(event.EventType)
    {
        case EET_KEY_INPUT_EVENT:
            return KeyEvent(event);

        case EET_MOUSE_INPUT_EVENT:
            return MouseEvent(event);

        case EET_GUI_EVENT:
            return GUIEvent(event);

        default:
            break;
    }

    return false;
}

void CPlayingState::RenderFrame(u32 cur_time)
{
    // Start by processing any input
    ProcessInput(cur_time);

    // Then process the trains
    m_train_manager.Process(cur_time);

    // Clear the screen
    ClearScreen();

    // Make sure that the scene manager draws its things
    GetDevice()->getSceneManager()->drawAll();
    GetDevice()->getGUIEnvironment()->drawAll();

    // End the scene
    Done();
}

bool CPlayingState::KeyEvent(const SEvent & event)
{
    // See which key was pushed.
    switch(event.KeyInput.Key)
    {
        case KEY_UP:
            m_up = event.KeyInput.PressedDown;
            break;

        case KEY_DOWN:
            m_down = event.KeyInput.PressedDown;
            break;

        case KEY_LEFT:
            m_left = event.KeyInput.PressedDown;
            break;

        case KEY_RIGHT:
            m_right = event.KeyInput.PressedDown;
            break;

        default:
            break;
    }

    // Now see if a key was let up
    if(!event.KeyInput.PressedDown)
    {
        if(event.KeyInput.Key == KEY_RETURN && m_mode == IM_PLACE_TRACK)
        {
            if(m_current_class == TC_STRAIGHT)
            {
                if(m_current_track == TT_VERTICAL_STRAIGHT)        m_current_track = TT_HORIZONTAL_STRAIGHT;
                else if(m_current_track == TT_HORIZONTAL_STRAIGHT) m_current_track = TT_VERTICAL_STRAIGHT;
            }
            if(m_current_class == TC_CURVE)
            {
                if(m_current_track == TT_CURVE_NE)      m_current_track = TT_CURVE_SE;
                else if(m_current_track == TT_CURVE_SE) m_current_track = TT_CURVE_SW;
                else if(m_current_track == TT_CURVE_SW) m_current_track = TT_CURVE_NW;
                else if(m_current_track == TT_CURVE_NW) m_current_track = TT_CURVE_NE;
            }
        }
    }

    return true;
}

bool CPlayingState::MouseEvent(const SEvent & event)
{
    // First, make sure that the coords
    // do not confilct with the GUI
    // Get the current values
    m_x_cur = event.MouseInput.X;
    m_y_cur = event.MouseInput.Y;

    IGUIElement * root = GetDevice()->getGUIEnvironment()->getRootGUIElement();

    if(root != root->getElementFromPoint(position2d<s32>(m_x_cur,m_y_cur)))
    {
        GetDevice()->getGUIEnvironment()->postEventFromUser(event);
        return true;
    }

    // Now see if any buttons were pressed
    // See which event was generated
    switch(event.MouseInput.Event)
    {
        case EMIE_LMOUSE_LEFT_UP:

            // Now we need to place the piece of track
            // at the current place.
            if(m_mode == IM_PLACE_TRACK)
            {
                int x = 0, z = 0;

                // Now place this piece
                if(m_ground.Intersection(GetDevice(), m_cursor_ray, x, z))
                    m_track_manager.SetPiece(m_current_track, x, z);
            }
            else if(m_mode == IM_PLACE_TRAIN)
            {
                int x = 0, z = 0;

                // Now place this train
                if(m_ground.Intersection(GetDevice(), m_cursor_ray, x, z))
                {
                    m_train_manager.AddTrain(GetDevice(), GetDevice()->getTimer()->getTime(), x, z);
                    m_mode = IM_SELECT;

                    UpdateGhostItems();
                }
            }

            break;

        case EMIE_RMOUSE_PRESSED_DOWN:

            // The button is down
            m_r_down = true;

            // Also reset mousemove values
            m_x_diff = m_y_diff = 0;
            break;

        case EMIE_RMOUSE_LEFT_UP:

            // Button is up now
            m_r_down = false;
            break;

        case EMIE_MOUSE_MOVED:

            // Add the new offsets to the existing ones
            m_x_diff += m_x_cur - m_x_prev;
            m_y_diff += m_y_cur - m_y_prev;

            // Now set the previous values
            m_x_prev = m_x_cur;
            m_y_prev = m_y_cur;

            // Given the cursor position, calculate a ray
            // that projects from the camera.
            m_cursor_ray = GetDevice()->getSceneManager()->
                                        getSceneCollisionManager()->
                                        getRayFromScreenCoordinates(position2d<s32>(m_x_cur,m_y_cur),
                                                                    NULL);

            break;

        case EMIE_MOUSE_WHEEL:

            {
                float fChange = event.MouseInput.Wheel * 0.1f;

                // Now calculate the amount to move the camera
                // along the vector defined by the camera to target
                vector3df v_target   = m_camera->getTarget();
                vector3df v_position = m_camera->getPosition();

                float x_total = v_position.X - v_target.X;
                float y_total = v_position.Y - v_target.Y;
                float z_total = v_position.Z - v_target.Z;

                // Calculate the amount to move along the vector
                v_position.X -= x_total * fChange;
                v_position.Y -= y_total * fChange;
                v_position.Z -= z_total * fChange;

                m_radius -= m_radius * fChange;

                m_camera->setPosition(v_position);
            }

            break;

        default:
            return false;
    }

    return true;
}

bool CPlayingState::GUIEvent(const SEvent & event)
{
    switch(event.GUIEvent.EventType)
    {
        // This will be the case for the toolbar buttons
        case EGET_BUTTON_CLICKED:

            if(event.GUIEvent.Caller->getID() == TOOLBAR_EDIT_TRACK)
            {
                if(!m_dialog_edittrack)
                    CreateTrackWindow();
                else
                {
                    // Close the track window
                    m_dialog_edittrack->remove();
                    m_dialog_edittrack = NULL;
                }
            }
            else if(event.GUIEvent.Caller->getID() == TOOLBAR_ADD_TRAIN)
            {
                if(!m_dialog_addtrain)
                    CreateTrainWindow();
                else
                {
                    m_dialog_addtrain->remove();
                    m_dialog_addtrain = NULL;
                }
            }
            else if(event.GUIEvent.Caller->getID() == EDITTRACK_STRAIGHT)
            {
                m_mode = IM_PLACE_TRACK;
                m_current_class = TC_STRAIGHT;
                m_current_track = TT_VERTICAL_STRAIGHT;

                UpdateGhostItems();
            }
            else if(event.GUIEvent.Caller->getID() == EDITTRACK_CURVED)
            {
                m_mode = IM_PLACE_TRACK;
                m_current_class = TC_CURVE;
                m_current_track = TT_CURVE_NE;

                UpdateGhostItems();
            }
            else if(event.GUIEvent.Caller->getID() == ADDTRAIN_ADD)
            {
                m_mode = IM_PLACE_TRAIN;
                UpdateGhostItems();

                // Close the dialog
                m_dialog_addtrain->remove();
                m_dialog_addtrain = NULL;

                // Switch off the train dialog
                IGUIButton * pButton = dynamic_cast<IGUIButton *>(m_toolbar->getElementFromId(TOOLBAR_ADD_TRAIN));
                pButton->setPressed(false);
            }

            break;

        case EGET_ELEMENT_CLOSED:

            // Set the mode and update the ghost items
            m_mode = IM_SELECT;
            UpdateGhostItems();

            // Depending on which dialog was closed, we
            // may need to unhighlight a button
            if(event.GUIEvent.Caller == m_dialog_edittrack)
            {
                IGUIButton * pButton = dynamic_cast<IGUIButton *>(m_toolbar->getElementFromId(TOOLBAR_EDIT_TRACK));
                pButton->setPressed(false);

                m_dialog_edittrack = NULL;
            }
            else if(event.GUIEvent.Caller == m_dialog_addtrain)
            {
                IGUIButton * pButton = dynamic_cast<IGUIButton *>(m_toolbar->getElementFromId(TOOLBAR_ADD_TRAIN));
                pButton->setPressed(false);

                m_dialog_addtrain = NULL;
            }

            return false;

        default:
            return false;
    }

    return true;
}

void CPlayingState::CreateToolbar()
{
    IGUIEnvironment* pGUI = GetDevice()->getGUIEnvironment();

    // Begin by creating the toolbar
    m_toolbar = pGUI->addToolBar();

    IGUIButton * pEditTrack = m_toolbar->addButton(TOOLBAR_EDIT_TRACK, NULL, NULL, GetDevice()->getVideoDriver()->getTexture(DATA_DIR "/icons/edit_track.png"), NULL, false, true);
    IGUIButton * pAddTrain = m_toolbar->addButton(TOOLBAR_ADD_TRAIN,  NULL, NULL, GetDevice()->getVideoDriver()->getTexture(DATA_DIR "/icons/add_train.png"), NULL, false, true);

    pEditTrack->setIsPushButton(true);
    pAddTrain->setIsPushButton(true);
}

void CPlayingState::CreateTrackWindow()
{
    // Create the track edit dialog
    m_dialog_edittrack = GetDevice()->getGUIEnvironment()->addWindow(rect<s32>(12,50,230,195), false, L"Edit Track");

    // Put the buttons in the window
    GetDevice()->getGUIEnvironment()->addStaticText(L"Track type:", rect<s32>(10,35,100,52), false, false, m_dialog_edittrack);
    IGUIButton * pStraight = GetDevice()->getGUIEnvironment()->addButton(rect<s32>(10,55,74,119), m_dialog_edittrack, EDITTRACK_STRAIGHT);
    IGUIButton * pCurved = GetDevice()->getGUIEnvironment()->addButton(rect<s32>(80,55,144,119), m_dialog_edittrack, EDITTRACK_CURVED);

    pStraight->setImage(GetDevice()->getVideoDriver()->getTexture(DATA_DIR "/icons/straight_track.png"));
    pCurved->setImage(GetDevice()->getVideoDriver()->getTexture(DATA_DIR "/icons/curved_track.png"));

    pStraight->setUseAlphaChannel(true);
    pCurved->setUseAlphaChannel(true);
}

void CPlayingState::CreateTrainWindow()
{
    // Show the add train dialog
    m_dialog_addtrain = GetDevice()->getGUIEnvironment()->addWindow(rect<s32>(12,50,230,195), false, L"Add Train");

    GetDevice()->getGUIEnvironment()->addStaticText(L"Engine:", rect<s32>(10,35,100,60), false, false, m_dialog_addtrain);

    // At this point, we only have one engine
    // so we only really need one entry in the
    // drop down box
    IGUIComboBox * pBox = GetDevice()->getGUIEnvironment()->addComboBox(rect<s32>(8,58,210,83), m_dialog_addtrain);

    pBox->addItem(L"Standard Diesel");

    // And add a button
    GetDevice()->getGUIEnvironment()->addButton(rect<s32>(8,100,210,126), m_dialog_addtrain, ADDTRAIN_ADD, L"Add train...");
}

void CPlayingState::UpdateGhostItems()
{
    m_straight_ghost_node->setVisible(m_mode == IM_PLACE_TRACK && m_current_class == TC_STRAIGHT);
    m_curve_ghost_node->setVisible(m_mode == IM_PLACE_TRACK && m_current_class == TC_CURVE);
    m_engine_ghost_node->setVisible(m_mode == IM_PLACE_TRAIN);
}

void CPlayingState::ProcessInput(u32 cur_time)
{
    // Get the difference from the previous time
    u32 diff = cur_time - m_previous_time;

    // Now get the current position of everything
    vector3df cur_pos = m_camera->getPosition();
    vector3df cur_tar = m_camera->getTarget();

    // Our rotation is increased everytime
    // the left / right arrow keys are used
    float fBefore = m_rotation;

    if(m_left)
        m_rotation += (diff * 0.002f);
    if(m_right)
        m_rotation -= (diff * 0.002f);

    // Calculate the cos/sin of the before values
    float bx = cosf(fBefore - PI2) * m_radius;
    float bz = sinf(fBefore - PI2) * m_radius;

    // Now for the new values
    float ax = cosf(m_rotation - PI2) * m_radius;
    float az = sinf(m_rotation - PI2) * m_radius;

    // Now get the difference
    cur_tar.X += bx - ax;
    cur_tar.Z += bz - az;

    if(m_r_down)
    {
        // When the mouse is moved, we track
        // the amount and move the target and
        // camera accordingly.

        // We need to make sure that the angle
        // of rotation (about the Y-Axis) is
        // taken into account here.

        // Precalculated numbers
        float xdiff = (m_x_diff * 0.01f * (m_radius / 3.0f));
        float zdiff = (m_y_diff * 0.02f * (m_radius / 3.0f));

        float xtrans = xdiff * cosf(m_rotation) + zdiff * sinf(m_rotation);
        float ztrans = zdiff * cosf(m_rotation) + xdiff * sinf(m_rotation - PI);

        cur_pos.X -= xtrans;
        cur_tar.X -= xtrans;
        cur_pos.Z += ztrans;
        cur_tar.Z += ztrans;

        m_x_diff = m_y_diff = 0;
    }

    m_camera->setPosition(cur_pos);
    m_camera->setTarget(cur_tar);

    // Deal with the track piece
    if(m_mode == IM_PLACE_TRACK)
    {
        int x = 0, z = 0;

        if(m_ground.Intersection(GetDevice(), m_cursor_ray, x, z))
        {
            // Depending on what the current piece is...
            float x_off = g_trackOrientationInfo[m_current_track].fTrans_x;
            float z_off = g_trackOrientationInfo[m_current_track].fTrans_z;

            float y_rot = g_trackOrientationInfo[m_current_track].fRot_y;

            if(m_current_track <= TT_HORIZONTAL_STRAIGHT)
            {
                m_straight_ghost_node->setPosition(vector3df(x * 2.0f + x_off, 0.05, z * 2.0f + z_off));
                m_straight_ghost_node->setRotation(vector3df(0,y_rot,0));
            }
            else
            {
                m_curve_ghost_node->setPosition(vector3df(x * 2.0f + x_off, 0.05, z * 2.0f + z_off));
                m_curve_ghost_node->setRotation(vector3df(0,y_rot,0));
            }
        }
    }
    else if(m_mode == IM_PLACE_TRAIN)
    {
        int x = 0, z = 0;

        if(m_ground.Intersection(GetDevice(), m_cursor_ray, x, z))
            m_engine_ghost_node->setPosition(vector3df(x * 2.0f + 1.0f, 0.05, z * 2.0f + 1.0f));
    }

    m_previous_time = cur_time;
}
