#include "CStateManager.h"

// State includes
#include "CLoadingState.h"
#include "CPlayingState.h"

CStateManager::CStateManager(IrrlichtDevice * pDevice)
    : m_device(pDevice), m_state(GS_LOADING), m_state_pointer(NULL), m_state_changing(false)
{
    // Grab a copy of the device
    m_device->grab();

    // At first, we will begin by showing the loading screen.
    SwitchState(GS_LOADING);
}

CStateManager::~CStateManager()
{
    // Make sure that the current state gets cleared
    if(m_state_pointer) delete m_state_pointer;

    // We're done with our copy of the device
    m_device->drop();
}

void CStateManager::Go()
{
    // This is the game's 'main loop'
    // Basically just continue as long as
    // run() is successful

    while(m_device->run())
    {
        u32 cur_time = m_device->getTimer()->getTime();

        // Delegate the rendering of the scene to
        // the current state
        m_state_pointer->RenderFrame(cur_time);

        // Now check to see if there are any pending tasks
        if(m_state_changing)
        {
            // Perform the state change
            SwitchState(m_state);

            m_state_changing = false;
        }
    }
}

void CStateManager::QueueStateChange(GAME_STATE new_state)
{
    m_state = new_state;

    // Note that we don't actually change the state
    // here, because this function may get called from
    // within the main loop and it would have the rug pulled
    // from right under it.

    // Instead, we set a flag that the main loop will recognize
    // and will then take action.
    m_state_changing = true;
}

void CStateManager::SwitchState(GAME_STATE new_state)
{
    m_state = new_state;

    // Free the old state
    if(m_state_pointer) delete m_state_pointer;

    // Now create the new state
    switch(m_state)
    {
        case GS_LOADING:
            m_state_pointer = new CLoadingState(this);
            break;

        case GS_PLAYING:
            m_state_pointer = new CPlayingState(this);
            break;
    }
}
