#include "CTrackManager.h"

// Definitions
const float PI2 = PI / 2;

CTrackManager::CTrackManager(IrrlichtDevice * pDevice, int width, int height)
    : m_width(width), m_height(height),
      m_straight(NULL), m_curved(NULL)
{
    // Temporarily set the current directory to the textures
    path cur_dir = pDevice->getFileSystem()->getWorkingDirectory();
    pDevice->getFileSystem()->changeWorkingDirectoryTo(DATA_DIR "/textures");

    // Start by loading all of the meshes
    m_straight = pDevice->getSceneManager()->getMesh("../models/track_straight.b3d");
    m_curved =   pDevice->getSceneManager()->getMesh("../models/track_small_curve.b3d");

    // Now change it back
    pDevice->getFileSystem()->changeWorkingDirectoryTo(cur_dir);

    // Grab a reference to these
    m_straight->grab();
    m_curved->grab();

    // Fill in the array of all the pieces
    for(int i=0;i<height;++i)
    {
        for(int j=0;j<width;++j)
        {
            // Calculate the position information for this
            // particular piece of track
            float piece_x = j * 2;
            float piece_z = i * 2;

            CTrack * pNew = new CTrack(pDevice,piece_x,0,piece_z,j,i);
            pNew->SetPiece(TT_NUM_PIECES,NULL);

            m_grid.push_back(pNew);
        }
    }
}

CTrackManager::~CTrackManager()
{
    // Free the meshes
    m_straight->drop();
    m_curved->drop();

    // Free all of the track pieces
    for(unsigned int i=0;i<m_grid.size();++i)
    {
        delete m_grid[i];
    }
}

void CTrackManager::SetPiece(TRACK_TYPE type,int x, int z)
{
    // Calculate this item's offset
    int iOffset = (z * m_width) + x;

    // Get that piece
    CTrack * piece = m_grid.at(iOffset);

    // Before we overwrite the current piece, we need
    // to make sure that the tile at the other end
    // of curved pieces is removed properly.
    TRACK_TYPE current_type = piece->GetPiece();

    if(current_type >= TT_CURVE_NE && current_type <= TT_CURVE_NW)
    {
        int iOtherOffset = iOffset;

        if(current_type == TT_CURVE_NE) { iOtherOffset -= 1; iOtherOffset += m_width; }
        if(current_type == TT_CURVE_SE) { iOtherOffset -= 1; iOtherOffset -= m_width; }
        if(current_type == TT_CURVE_SW) { iOtherOffset += 1; iOtherOffset -= m_width; }
        if(current_type == TT_CURVE_NW) { iOtherOffset += 1; iOtherOffset += m_width; }

        CTrack * other_piece = m_grid.at(iOtherOffset);
        other_piece->SetPiece(TT_NUM_PIECES, NULL);
    }

    // AND make sure that if this is the other
    // end of a curved piece that we set the other end
    // to nothing
    if(current_type >= TT_CURVE_NE_OTHER && current_type <= TT_CURVE_NW_OTHER)
    {
        int iOtherOffset = iOffset;

        if(current_type == TT_CURVE_NE_OTHER) { iOtherOffset += 1; iOtherOffset -= m_width; }
        if(current_type == TT_CURVE_SE_OTHER) { iOtherOffset += 1; iOtherOffset += m_width; }
        if(current_type == TT_CURVE_SW_OTHER) { iOtherOffset -= 1; iOtherOffset += m_width; }
        if(current_type == TT_CURVE_NW_OTHER) { iOtherOffset -= 1; iOtherOffset -= m_width; }

        CTrack * other_piece = m_grid.at(iOtherOffset);
        other_piece->SetPiece(TT_NUM_PIECES, NULL);
    }

    // Now set that piece
    if(type < TT_CURVE_NE)
        piece->SetPiece(type,m_straight);
    else
    {
        piece->SetPiece(type, m_curved);
        // For the curved pieces, we need to
        // also set the tile at the other end to
        // the _OTHER track piece
        int iOtherOffset = iOffset;
        TRACK_TYPE other_type;

        if(type == TT_CURVE_NE) { iOtherOffset -= 1; iOtherOffset += m_width; other_type = TT_CURVE_NE_OTHER; }
        if(type == TT_CURVE_SE) { iOtherOffset -= 1; iOtherOffset -= m_width; other_type = TT_CURVE_SE_OTHER; }
        if(type == TT_CURVE_SW) { iOtherOffset += 1; iOtherOffset -= m_width; other_type = TT_CURVE_SW_OTHER; }
        if(type == TT_CURVE_NW) { iOtherOffset += 1; iOtherOffset += m_width; other_type = TT_CURVE_NW_OTHER; }

        CTrack * other_piece = m_grid.at(iOtherOffset);
        other_piece->SetPiece(other_type, NULL);
    }
}

CTrack * CTrackManager::GetPiece(int x, int z)
{
    // Calculate this item's offset
    int iOffset = (z * m_width) + x;

    // Get that piece
    CTrack * piece = m_grid.at(iOffset);

    // Return the type
    return piece;
}

void CTrackManager::CalculateTrainPosition(CTrack * pTrack, DIRECTION dir, float fPos, float & fx, float & fz, float & fy)
{
    // Determine what type of track
    // we are dealing with

    switch(pTrack->GetPiece())
    {
        case TT_VERTICAL_STRAIGHT:

            // The translation and rotation for the track piece
            fx = 1.0f;
            if(dir == D_NORTH) { fy =   0.0; fz = fPos * 2.0f; }
            else               { fy = 180.0; fz = 2.0f - (fPos * 2.0f); }

            break;

        case TT_HORIZONTAL_STRAIGHT:

            // The translation and rotation for the track piece
            fz = 1.0f;
            if(dir == D_EAST) { fy =  90.0; fx = fPos * 2.0f; }
            else              { fy = 270.0; fx = 2.0 - (fPos * 2.0f); }

            break;

        case TT_CURVE_NE:

            // The translation for the track
            fx = (cosf(fPos * PI2) * 3.0f) - 2.0f;
            fz = (sinf(fPos * PI2) * 3.0f);

            // The rotation
            fy = -(90.0 * fPos);

            break;

        case TT_CURVE_SE:

            // The translation for the track
            fx = (cosf(fPos * PI2) * 3.0f) - 2.0f;
            fz = (-sinf(fPos * PI2) * 3.0f) + 2.0f;

            // The rotation
            fy = (90.0 * fPos) - 180.0f;

            break;

        case TT_CURVE_SW:

            // The translation for the track
            fx = (-cosf(fPos * PI2) * 3.0f) + 4.0f;
            fz = (-sinf(fPos * PI2) * 3.0f) + 2.0f;

            // The rotation
            fy = 180.0f - (90.0f * fPos);

            break;

        case TT_CURVE_NW:

            // The translation for the track
            fx = (-cosf(fPos * PI2) * 3.0f) + 4.0f;
            fz = (sinf(fPos * PI2) * 3.0f);

            // The rotation
            fy = (90.0 * fPos);

            break;

        case TT_CURVE_NE_OTHER:

            // The translation for the track
            fx = (-cosf(fPos * PI2 + PI2) * 3.0f);
            fz = (sinf(fPos * PI2 + PI2) * 3.0f) - 2.0f;

            // The rotation
            fy = (90.0 * fPos) + 90.0f;

            break;

        case TT_CURVE_SE_OTHER:

            // The translation for the track
            fx = (-cosf(fPos * PI2 + PI2) * 3.0f);
            fz = (sinf(fPos * PI2 - PI2) * 3.0f) + 4.0f;

            // The rotation
            fy = 90.0f - (90.0 * fPos);

            break;

        case TT_CURVE_SW_OTHER:

            // The translation for the track
            fx = (cosf(fPos * PI2 + PI2) * 3.0f) + 2.0f;
            fz = (sinf(fPos * PI2 - PI2) * 3.0f) + 4.0f;

            // The rotation
            fy = (90.0 * fPos) - 90;

            break;

        case TT_CURVE_NW_OTHER:

            // The translation for the track
            fx = (cosf(fPos * PI2 + PI2) * 3.0f) + 2.0f;
            fz = (sinf(fPos * PI2 + PI2) * 3.0f) - 2.0f;

            // The rotation
            fy = -(90.0 * fPos) - 90.0f;

            break;

        default:
            break;
    }
}

CTrack * CTrackManager::GetNextTrack(CTrack * pCurrent, DIRECTION d_current, DIRECTION & d_new)
{
    // Calculate the index into the current piece
    int iOffset = (pCurrent->GetZOffset() * m_width) + pCurrent->GetXOffset();

    // Determine the next piece of track
    TRACK_TYPE cur_type = pCurrent->GetPiece();

    switch(cur_type)
    {
        case TT_VERTICAL_STRAIGHT:

            if(d_current == D_NORTH) iOffset += m_width;
            else                     iOffset -= m_width;

            // Direction is unchanged
            d_new = d_current;

            break;

        case TT_HORIZONTAL_STRAIGHT:

            if(d_current == D_EAST) iOffset += 1;
            else                    iOffset -= 1;

            // Direction is unchanged
            d_new = d_current;

            break;

        // The curved pieces
        case TT_CURVE_NE:

            iOffset += g_trackOrientationInfo[cur_type].iOffsetX;
            iOffset += g_trackOrientationInfo[cur_type].iOffsetZ * m_width;
            d_new = D_WEST;
            break;

        case TT_CURVE_SE:

            iOffset += g_trackOrientationInfo[cur_type].iOffsetX;
            iOffset += g_trackOrientationInfo[cur_type].iOffsetZ * m_width;
            d_new = D_WEST;
            break;

        case TT_CURVE_SW:

            iOffset += g_trackOrientationInfo[cur_type].iOffsetX;
            iOffset += g_trackOrientationInfo[cur_type].iOffsetZ * m_width;
            d_new = D_EAST;
            break;

        case TT_CURVE_NW:

            iOffset += g_trackOrientationInfo[cur_type].iOffsetX;
            iOffset += g_trackOrientationInfo[cur_type].iOffsetZ * m_width;
            d_new = D_EAST;
            break;

        // The other end of the curved pieces
        case TT_CURVE_NE_OTHER:

            iOffset += g_trackOrientationInfo[cur_type].iOffsetX;
            iOffset += g_trackOrientationInfo[cur_type].iOffsetZ * m_width;
            d_new = D_SOUTH;
            break;

        case TT_CURVE_SE_OTHER:

            iOffset += g_trackOrientationInfo[cur_type].iOffsetX;
            iOffset += g_trackOrientationInfo[cur_type].iOffsetZ * m_width;
            d_new = D_NORTH;
            break;

        case TT_CURVE_SW_OTHER:

            iOffset += g_trackOrientationInfo[cur_type].iOffsetX;
            iOffset += g_trackOrientationInfo[cur_type].iOffsetZ * m_width;
            d_new = D_NORTH;
            break;

        case TT_CURVE_NW_OTHER:

            iOffset += g_trackOrientationInfo[cur_type].iOffsetX;
            iOffset += g_trackOrientationInfo[cur_type].iOffsetZ * m_width;
            d_new = D_SOUTH;
            break;

        default:
            break;
    }

    return m_grid.at(iOffset);
}
